/*
** File: admxrc6tx_common.c  
** Project: ADB3 core driver
** Purpose: ADM-XRC-6TL/1-specific code for various hardware functions.
**
** (C) Copyright Alpha Data 2010
*/

#include <df.h>

#include "adb3.h"
#include "admxrc6tx_common.h"

typedef struct _SensorInfo {
  const char* pDescription;
  uint32_t capabilities;
  struct {
    int16_t intPart;
    uint16_t fracPart;
  } error;
  uint32_t datatype;
  uint32_t unit;
  int32_t exponent;
} SensorInfo;

static SensorInfo g_lm87SensorInfo[] = {
  /* LM87 sensors */
  { "1V supply rail",              0, { 0,  1311 }, CoreSensorDataDouble, CoreSensorUnitV,    0 }, /* LM87 register 0x21, Vccp1 reading */
  { "1.5V supply rail",            0, { 0,  1966 }, CoreSensorDataDouble, CoreSensorUnitV,    0 }, /* LM87 register 0x28, FAN1/AIN1 reading */
  { "1.8V supply rail",            0, { 0,  2359 }, CoreSensorDataDouble, CoreSensorUnitV,    0 }, /* LM87 register 0x29, FAN2/AIN2 reading */
  { "2.5V supply rail",            0, { 0,  3277 }, CoreSensorDataDouble, CoreSensorUnitV,    0 }, /* LM87 register 0x20, +2.5V / ext temp. 2 reading */
  { "3.3V supply rail",            0, { 0,  4325 }, CoreSensorDataDouble, CoreSensorUnitV,    0 }, /* LM87 register 0x22, +Vcc reading */
  { "5V supply rail",              0, { 0,  6554 }, CoreSensorDataDouble, CoreSensorUnitV,    0 }, /* LM87 register 0x23, 5V reading */
  { "XMC variable power rail",     0, { 0, 15729 }, CoreSensorDataDouble, CoreSensorUnitV,    0 }, /* LM87 register 0x24, +12V reading */
  { "XRM I/O voltage",             0, { 0,  3277 }, CoreSensorDataDouble, CoreSensorUnitV,    0 }, /* LM87 register 0x25, Vccp2 reading */
  { "LM87 internal temperature",   0, { 3,     0 }, CoreSensorDataDouble, CoreSensorUnitC,    0 }, /* LM87 register 0x27, Int temp reading */
  { "Target FPGA ext. temp.",      0, { 4,     0 }, CoreSensorDataDouble, CoreSensorUnitC,    0 }  /* LM87 register 0x26, Ext temp. 1 reading */
};

/* These correspond to the elements of 'g_sensorInfo' above, to convert the raw ADC values into Volt / deg. C */
static struct {
  unsigned int regOffset; /* LM87 register offset */
  int32_t factor;         /* Fixed point, where 0x00100000 => 1, 0x00080000 => 0.5, 0xFFF80000 => -0.5 etc. */
} g_lm87Scaling[] = {
  { 0x21U, 14785    /* 0.0141f */ },
  { 0x28U, 10276,   /* 0.0098f */ },
  { 0x29U, 10276,   /* 0.0098f */ },
  { 0x20U, 13631,   /* 0.0130f */ },
  { 0x22U, 18036,   /* 0.0172f */ },
  { 0x23U, 27263,   /* 0.0260f */ },
  { 0x24U, 65536,   /* 0.0625f */ },
  { 0x25U, 14785,   /* 0.0141f */ },
  { 0x27U, 1048576, /* 1.0000f */ },
  { 0x26U, 1048576, /* 1.0000f */ }
};

/* Xilinx System Monitor sensors, available with firmware 1.4 (PCI rev ID 0x04) and later */
static SensorInfo g_xsmonSensorInfo[] = {
  { "Bridge temperature",          0, { 5,     0 }, CoreSensorDataDouble, CoreSensorUnitC,    0 }, /* XSMON internal FPGA temp. reading */
  { "Bridge VCCINT",               0, { 0,  3932 }, CoreSensorDataDouble, CoreSensorUnitV,    0 }, /* XSMON internal FPGA VINT reading */
  { "Bridge VCCAUX",               0, { 0,  3932 }, CoreSensorDataDouble, CoreSensorUnitV,    0 }  /* XSMON internal FPGA VAUX reading */
};

/* These correspond to the elements of 'g_xsmonSensorInfo' above, to convert the raw ADC values into Volt / deg. C */
static struct {
  int32_t factor;
  int32_t offset;
} g_xsmonScaling[] = {
  { 32254, -17901158 }, /* T(C)  = x * (503.975 / 1024) - 273.15 */
  { 192,    0 },        /* V(mV) = x * (3 / 1024) */
  { 192,    0 }         /* V(mV) = x * (3 / 1024) */
};

/* Extra diagnostic sensors that have to be enabled by setting driver parameter EnableDiagSensor to TRUE */
static SensorInfo g_diagSensorInfo[] = {
  /* Sensors that are fields in ADB3 PWRSTAT register */
  { "1.5V & 2.5V supply good",     0, { 0,     0 }, CoreSensorDataBool,   CoreSensorUnitNone, 0 }, /* ADB3 PWRSTAT.PSU1_GOOD */
  { "1.0V supply good",            0, { 0,     0 }, CoreSensorDataBool,   CoreSensorUnitNone, 0 }, /* ADB3 PWRSTAT.PSU2_GOOD */
  { "1.8V supply good",            0, { 0,     0 }, CoreSensorDataBool,   CoreSensorUnitNone, 0 }, /* ADB3 PWRSTAT.DIG_1V8_OK */
  { "DDR-3 SDRAM VRef A good",     0, { 0,     0 }, CoreSensorDataBool,   CoreSensorUnitNone, 0 }, /* ADB3 PWRSTAT.VREF_DRAM_A_OK */
  { "DDR-3 SDRAM VRef B good",     0, { 0,     0 }, CoreSensorDataBool,   CoreSensorUnitNone, 0 }, /* ADB3 PWRSTAT.VREF_DRAM_B_OK */
  { "Bridge MGT 1.0V supply good", 0, { 0,     0 }, CoreSensorDataBool,   CoreSensorUnitNone, 0 }, /* ADB3 PWRSTAT.BRG_AVCC_1V0_OK */
  { "Bridge MGT 1.2V supply good", 0, { 0,     0 }, CoreSensorDataBool,   CoreSensorUnitNone, 0 }, /* ADB3 PWRSTAT.BRG_AVCC_1V2_OK */
  { "Target MGT 1.0V supply good", 0, { 0,     0 }, CoreSensorDataBool,   CoreSensorUnitNone, 0 }, /* ADB3 PWRSTAT.TGT_AVCC_1V0_OK */
  { "Target MGT 1.2V supply good", 0, { 0,     0 }, CoreSensorDataBool,   CoreSensorUnitNone, 0 }, /* ADB3 PWRSTAT.TGT_AVCC_1V2_OK */
  { "XRM VI/O supply good",        0, { 0,     0 }, CoreSensorDataBool,   CoreSensorUnitNone, 0 }, /* ADB3 PWRSTAT.XRM_VIO_OK */
  { "XRM VI/O select lines [2:0]", 0, { 0,     0 }, CoreSensorDataUInt32, CoreSensorUnitNone, 0 }, /* ADB3 PWRSTAT.XRMVSEL */
  { "XRM VI/O supply enabled",     0, { 0,     0 }, CoreSensorDataBool,   CoreSensorUnitNone, 0 }, /* ADB3 PWRSTAT.XRMVEN */
  { "XRM presence detect",         0, { 0,     0 }, CoreSensorDataBool,   CoreSensorUnitNone, 0 }, /* ADB3 PWRSTAT.XRM_PRESENT */
  { "XRM force 2.5V",              0, { 0,     0 }, CoreSensorDataBool,   CoreSensorUnitNone, 0 }, /* ADB3 PWRSTAT.FORCE2V5 */
  { "XRM ROM VI/O(Nom) valid",     0, { 0,     0 }, CoreSensorDataBool,   CoreSensorUnitNone, 0 }, /* ADB3 PWRSTAT.NOMV_VD */
  { "XRM ROM VI/O(Nom) OOB",       0, { 0,     0 }, CoreSensorDataBool,   CoreSensorUnitNone, 0 }, /* ADB3 PWRSTAT.OOB_VREQ */
  { "XRM ROM error",               0, { 0,     0 }, CoreSensorDataBool,   CoreSensorUnitNone, 0 }, /* ADB3 PWRSTAT.ERROR */
  /* Sensors that are fields in ADB3 PWRSTAT register */
  { "Target FPGA PROG# asserted",  0, { 0,     0 }, CoreSensorDataBool,   CoreSensorUnitNone, 0 }, /* ADB3 FPCTL.PROG */
  { "Target FPGA INIT# asserted",  0, { 0,     0 }, CoreSensorDataBool,   CoreSensorUnitNone, 0 }, /* ADB3 FPCTL.INIT */
  { "Target FPGA DONE asserted",   0, { 0,     0 }, CoreSensorDataBool,   CoreSensorUnitNone, 0 }, /* ADB3 FPCTL.DONE */
  { "Target FPGA BUSY asserted",   0, { 0,     0 }, CoreSensorDataBool,   CoreSensorUnitNone, 0 }, /* ADB3 FPCTL.BUSY */
  { "Target FPGA MODE[2:0]",       0, { 0,     0 }, CoreSensorDataUInt32, CoreSensorUnitNone, 0 }, /* ADB3 FPCTL.MODE */
  { "Target FPGA FBS# asserted",   0, { 0,     0 }, CoreSensorDataBool,   CoreSensorUnitNone, 0 }, /* ADB3 FPCTL.FBS */
  { "Target FPGA ONECONFIG",       0, { 0,     0 }, CoreSensorDataBool,   CoreSensorUnitNone, 0 }, /* ADB3 FPCTL.ONECONFIG */
  { "Target FPGA INDGRADE",        0, { 0,     0 }, CoreSensorDataBool,   CoreSensorUnitNone, 0 }, /* ADB3 FPCTL.INDGRADE */
  { "Target FPGA TGT_CLEARED",     0, { 0,     0 }, CoreSensorDataBool,   CoreSensorUnitNone, 0 }, /* ADB3 FPCTL.TGT_CLEARED */
  { "Target FPGA RELOAD",          0, { 0,     0 }, CoreSensorDataBool,   CoreSensorUnitNone, 0 }, /* ADB3 FPCTL.RELOAD */
  /* Sensors that are fields in ADB3 MPTL_RECOVER register */
  { "MPTL_RECOVER.AR_COUNT",       0, { 0,     0 }, CoreSensorDataUInt32, CoreSensorUnitNone, 0 }, /* ADB3 MPTL_RECOVER.AR_COUNT */
  { "MPTL_RECOVER.OCP_INFLIGHT",   0, { 0,     0 }, CoreSensorDataUInt32, CoreSensorUnitNone, 0 }, /* ADB3 MPTL_RECOVER.OCP_RECOVERY_INFLIGHT */
  /* Sensors that are fields in ADB3 MPTL_STAT register */
  { "MPTL_STAT.TARGET_CONFIGURED", 0, { 0,     0 }, CoreSensorDataBool,   CoreSensorUnitNone, 0 }, /* ADB3 MPTL_STAT.TARGET_CONFIGURED */
  { "MPTL_STAT.LOCAL_GTX_ONLINE",  0, { 0,     0 }, CoreSensorDataBool,   CoreSensorUnitNone, 0 }, /* ADB3 MPTL_STAT.LOCAL_GTX_ONLINE */
  { "MPTL_STAT.TARGET_GTX_ONLINE", 0, { 0,     0 }, CoreSensorDataBool,   CoreSensorUnitNone, 0 }, /* ADB3 MPTL_STAT.TARGET_GTX_ONLINE */
  { "MPTL_STAT.GTX_CLK_LOCKED",    0, { 0,     0 }, CoreSensorDataBool,   CoreSensorUnitNone, 0 }, /* ADB3 MPTL_STAT.GTX_CLK_LOCKED */
  { "MPTL_STAT.SYNC_LOSS_COUNT",   0, { 0,     0 }, CoreSensorDataUInt32, CoreSensorUnitNone, 0 }  /* ADB3 MPTL_STAT.SYNC_LOSS_COUNT */
};

static struct {
  unsigned int offset;
  unsigned int startBit;
  unsigned int length;
} g_diagSensorFields[] = {
  /* PWRSTAT fields */
  { DF_OFFSET_OF(ModelRegsAdmxrc6tx, pwrStat),       0, 1 }, 
  { DF_OFFSET_OF(ModelRegsAdmxrc6tx, pwrStat),       1, 1 }, 
  { DF_OFFSET_OF(ModelRegsAdmxrc6tx, pwrStat),       2, 1 }, 
  { DF_OFFSET_OF(ModelRegsAdmxrc6tx, pwrStat),       3, 1 }, 
  { DF_OFFSET_OF(ModelRegsAdmxrc6tx, pwrStat),       4, 1 }, 
  { DF_OFFSET_OF(ModelRegsAdmxrc6tx, pwrStat),       5, 1 }, 
  { DF_OFFSET_OF(ModelRegsAdmxrc6tx, pwrStat),       6, 1 }, 
  { DF_OFFSET_OF(ModelRegsAdmxrc6tx, pwrStat),       7, 1 }, 
  { DF_OFFSET_OF(ModelRegsAdmxrc6tx, pwrStat),       8, 1 }, 
  { DF_OFFSET_OF(ModelRegsAdmxrc6tx, pwrStat),       9, 1 }, 
  { DF_OFFSET_OF(ModelRegsAdmxrc6tx, pwrStat),      16, 3 }, 
  { DF_OFFSET_OF(ModelRegsAdmxrc6tx, pwrStat),      19, 1 }, 
  { DF_OFFSET_OF(ModelRegsAdmxrc6tx, pwrStat),      24, 1 }, 
  { DF_OFFSET_OF(ModelRegsAdmxrc6tx, pwrStat),      25, 1 }, 
  { DF_OFFSET_OF(ModelRegsAdmxrc6tx, pwrStat),      26, 1 }, 
  { DF_OFFSET_OF(ModelRegsAdmxrc6tx, pwrStat),      27, 1 }, 
  { DF_OFFSET_OF(ModelRegsAdmxrc6tx, pwrStat),      31, 1 }, 
  /* FPCTL fields */
  { DF_OFFSET_OF(ModelRegsAdmxrc6tx, fpgaCtl),       0, 1 }, 
  { DF_OFFSET_OF(ModelRegsAdmxrc6tx, fpgaCtl),       1, 1 }, 
  { DF_OFFSET_OF(ModelRegsAdmxrc6tx, fpgaCtl),       2, 1 }, 
  { DF_OFFSET_OF(ModelRegsAdmxrc6tx, fpgaCtl),       3, 1 }, 
  { DF_OFFSET_OF(ModelRegsAdmxrc6tx, fpgaCtl),       8, 3 }, 
  { DF_OFFSET_OF(ModelRegsAdmxrc6tx, fpgaCtl),      27, 1 }, 
  { DF_OFFSET_OF(ModelRegsAdmxrc6tx, fpgaCtl),      28, 1 }, 
  { DF_OFFSET_OF(ModelRegsAdmxrc6tx, fpgaCtl),      29, 1 },
  { DF_OFFSET_OF(ModelRegsAdmxrc6tx, fpgaCtl),      30, 1 },
  { DF_OFFSET_OF(ModelRegsAdmxrc6tx, fpgaCtl),      31, 1 },
  /* MPTL_RECOVER fields */
  { DF_OFFSET_OF(ModelRegsAdmxrc6tx, mptlRecovery),  4, 4 },
  { DF_OFFSET_OF(ModelRegsAdmxrc6tx, mptlRecovery),  8, 8 },
  /* MPTL_STAT fields */
  { DF_OFFSET_OF(ModelRegsAdmxrc6tx, mptlStat),      0, 1 },
  { DF_OFFSET_OF(ModelRegsAdmxrc6tx, mptlStat),      1, 1 },
  { DF_OFFSET_OF(ModelRegsAdmxrc6tx, mptlStat),      2, 1 },
  { DF_OFFSET_OF(ModelRegsAdmxrc6tx, mptlStat),      3, 1 },
  { DF_OFFSET_OF(ModelRegsAdmxrc6tx, mptlStat),      8, 24 }
};

unsigned int g_admxrc6txNumDiagSensor = DF_ARRAY_LENGTH(g_diagSensorInfo);

boolean_t
admxrc6txIdentifyXrm(
	Adb3CoreDeviceContext* pDevCtx)
{
  CoreIoModuleInfo* pInfo = &pDevCtx->info.ioModule[0];
	ModelRegsAdmxrc6tx* pModelRegs;
  DfMemoryHandle hModel;
  boolean_t bAssumeXrmPresent = FALSE, bSuccess, bForce2V5;
  uint8_t val8;
  uint32_t pwrStat;
  unsigned int i;

  pModelRegs = pDevCtx->hardware.model.pAdmxrc6tx;
  hModel = pDevCtx->hardware.hModel;

  /* This driver parameter can be used to override presence detection for XRMs that don't drive the presence pin correctly */
  dfParameterGetBoolean(pDevCtx->pDevObj->pDrvObj, "AssumeXrmPresent", &bAssumeXrmPresent);

  pwrStat = dfPciMemRead32(hModel, &pModelRegs->pwrStat);
  pInfo->bPresent = (pwrStat & ADMXRC6Tx_V_PWRSTAT_XPRSNT) == ADMXRC6Tx_V_PWRSTAT_XPRSNT ? TRUE : FALSE;
  if (bAssumeXrmPresent) {
    pInfo->bPresent = TRUE;
  }
  bForce2V5 = (pwrStat & ADMXRC6Tx_V_PWRSTAT_XFORCE2V5) == ADMXRC6Tx_V_PWRSTAT_XFORCE2V5 ? TRUE : FALSE;
  bSuccess = admxrc6txReadXrmByte(pDevCtx, 0, &val8);
  if (bSuccess && val8 == 0x1U) {
    if (pInfo->bPresent) {
      dfDebugPrint(2, ("identifyXrm: XRM has FRU ROM\n"));
    } else {
      /* Well behaved XRMs should assert the presence pin, but some don't */
      pInfo->bPresent = TRUE;
      dfDebugPrint(2, ("identifyXrm: XRM has FRU ROM, but doesn't assert presence pin\n"));
    }
    pInfo->pVpdBuffer = dfMalloc(ADMXRC6TX_FRU_ROM_SIZE);
    if (NULL == pInfo->pVpdBuffer) {
      dfDebugPrint(0, ("*** identifyXrm: Failed to allocate FRU ROM buffer, size %p bytes\n", (void*)(uintptr_t)ADMXRC6TX_FRU_ROM_SIZE));
      return FALSE;
    }
    pInfo->vpdLength = ADMXRC6TX_FRU_ROM_SIZE;
    pInfo->bLegacyXrm = FALSE;
    pInfo->bForce2v5 = bForce2V5;
    pInfo->pVpdBuffer[0] = val8;
    for (i = 1; i < ADMXRC6TX_FRU_ROM_SIZE; i++) {
      bSuccess = admxrc6txReadXrmByte(pDevCtx, (uint8_t)i, &pInfo->pVpdBuffer[i]);
      if (!bSuccess) {
        dfDebugPrint(0, ("*** identifyXrm: Failed to read byte from FRU ROM 0x%x\n", i));
        return FALSE;
      }
    }
  } else {
    if (pInfo->bPresent) {
      dfDebugPrint(2, ("identifyXrm: FRU ROM not detected - assuming legacy XRM\n"));
      pInfo->bLegacyXrm = TRUE;
      pInfo->bForce2v5 = bForce2V5;
    } else {
      dfDebugPrint(2, ("identifyXrm: XRM not present\n"));
      pInfo->bPresent = FALSE;
    }
  }

  return TRUE;
}

void
admxrc6txInitSensorInfo(
	Adb3CoreDeviceContext* pDevCtx)
{
  unsigned int i, j = 0;

  dfAssert(j + DF_ARRAY_LENGTH(g_lm87SensorInfo) <= MAX_NUM_SENSOR);
  for (i = 0; i < DF_ARRAY_LENGTH(g_lm87SensorInfo) && j < MAX_NUM_SENSOR; i++, j++) {
    dfStrCpy(pDevCtx->info.sensor[j].description, DF_ARRAY_LENGTH(pDevCtx->info.sensor[j].description), g_lm87SensorInfo[i].pDescription);
    pDevCtx->info.sensor[j].capabilities = g_lm87SensorInfo[i].capabilities;
    pDevCtx->info.sensor[j].error.intPart = g_lm87SensorInfo[i].error.intPart;
    pDevCtx->info.sensor[j].error.fracPart = g_lm87SensorInfo[i].error.fracPart;
    pDevCtx->info.sensor[j].datatype = g_lm87SensorInfo[i].datatype;
    pDevCtx->info.sensor[j].unit = g_lm87SensorInfo[i].unit;
    pDevCtx->info.sensor[j].exponent = g_lm87SensorInfo[i].exponent;
  }
  if (pDevCtx->pDevObj->hardwareId.variant.pci.revision >= 0x04U) {
    for (i = 0; i < DF_ARRAY_LENGTH(g_xsmonSensorInfo) && j < MAX_NUM_SENSOR; i++, j++) {
      dfStrCpy(pDevCtx->info.sensor[j].description, DF_ARRAY_LENGTH(pDevCtx->info.sensor[j].description), g_xsmonSensorInfo[i].pDescription);
      pDevCtx->info.sensor[j].capabilities = g_xsmonSensorInfo[i].capabilities;
      pDevCtx->info.sensor[j].error.intPart = g_xsmonSensorInfo[i].error.intPart;
      pDevCtx->info.sensor[j].error.fracPart = g_xsmonSensorInfo[i].error.fracPart;
      pDevCtx->info.sensor[j].datatype = g_xsmonSensorInfo[i].datatype;
      pDevCtx->info.sensor[j].unit = g_xsmonSensorInfo[i].unit;
      pDevCtx->info.sensor[j].exponent = g_xsmonSensorInfo[i].exponent;
    }
  }
  if (pDevCtx->info.bootstrap.bIncludeDiagSensors) {
    dfAssert(j + DF_ARRAY_LENGTH(g_diagSensorInfo) <= MAX_NUM_SENSOR);
    for (i = 0; i < DF_ARRAY_LENGTH(g_diagSensorInfo) && j < MAX_NUM_SENSOR; i++, j++) {
      dfStrCpy(pDevCtx->info.sensor[j].description, DF_ARRAY_LENGTH(pDevCtx->info.sensor[j].description), g_diagSensorInfo[i].pDescription);
      pDevCtx->info.sensor[j].capabilities = g_diagSensorInfo[i].capabilities;
      pDevCtx->info.sensor[j].error.intPart = g_diagSensorInfo[i].error.intPart;
      pDevCtx->info.sensor[j].error.fracPart = g_diagSensorInfo[i].error.fracPart;
      pDevCtx->info.sensor[j].datatype = g_diagSensorInfo[i].datatype;
      pDevCtx->info.sensor[j].unit = g_diagSensorInfo[i].unit;
      pDevCtx->info.sensor[j].exponent = g_diagSensorInfo[i].exponent;
    }
  }
}

CoreSensorStatus
admxrc6txReadSensor(
	Adb3CoreDeviceContext* pDevCtx,
  unsigned int index,
  CoreSensorValue* pReading)
{
  ModelRegsAdmxrc6tx* pModelRegs;
  DfMemoryHandle hModel;
  uint32_t val32;
  int16_t val16;
  int32_t factor, offset, value;
  uint8_t* pReg;

  if (index >= pDevCtx->info.bootstrap.numSensor) {
    return CoreSensorInvalidIndex;
  }

  pModelRegs = pDevCtx->hardware.model.pAdmxrc6tx;
  hModel = pDevCtx->hardware.hModel;

  dfAssert(DF_ARRAY_LENGTH(g_lm87SensorInfo) == DF_ARRAY_LENGTH(g_lm87Scaling));
  if (index < DF_ARRAY_LENGTH(g_lm87SensorInfo)) {
    /* LM87 sensors */
    val16 = (int16_t)dfPciMemRead8(hModel, pModelRegs->sysMonBuf + g_lm87Scaling[index].regOffset);
    factor = g_lm87Scaling[index].factor;
    value = (val16 * factor) >> 4;
    pReading->doubleValue.intPart = (int16_t)(value >> 16);
    pReading->doubleValue.fracPart = (uint16_t)value;
    return CoreSensorSuccess;
  }
  index -= DF_ARRAY_LENGTH(g_lm87SensorInfo);

  if (pDevCtx->pDevObj->hardwareId.variant.pci.revision >= 0x04U) {
    if (index < DF_ARRAY_LENGTH(g_xsmonSensorInfo)) {
      uint32_t* pRegister;
      uint16_t val10;

      pRegister = (uint32_t*)&pModelRegs->xsmonTemp1 + (index << 1);
      /* Get 10-bit A/D reading */
      val10 = (uint16_t)((dfPciMemRead32(hModel, pRegister) >> 6) & 0x3ffU);
      /* Get scaling factor and offset in i15.u16 fixed-point format */
      factor = g_xsmonScaling[index].factor;
      offset = g_xsmonScaling[index].offset;
      /* Apply formula using i15.u16 fixed-point arithmetic */
      value = (int32_t)val10 * factor + offset;
      pReading->doubleValue.intPart = (int16_t)(value >> 16);
      pReading->doubleValue.fracPart = (uint16_t)value;
      return CoreSensorSuccess;
    }
    index -= DF_ARRAY_LENGTH(g_xsmonSensorInfo);
  }

  if (index < DF_ARRAY_LENGTH(g_diagSensorInfo)) {
    dfAssert(DF_ARRAY_LENGTH(g_diagSensorInfo) == DF_ARRAY_LENGTH(g_diagSensorFields));
    pReg = (uint8_t*)pModelRegs + g_diagSensorFields[index].offset;
    val32 = dfPciMemRead32(hModel, (uint32_t*)pReg);
    val32 >>= g_diagSensorFields[index].startBit;
    val32 &= DF_MASK32U(g_diagSensorFields[index].length);
    switch (g_diagSensorInfo[index].datatype) {
    case CoreSensorDataBool:
      pReading->booleanValue = val32 ? TRUE : FALSE;
      break;

    case CoreSensorDataUInt32:
      pReading->uint32Value = val32;
      break;

    case CoreSensorDataInt32:
      pReading->int32Value = (int32_t)val32;
      break;

    case CoreSensorDataDouble:
      pReading->doubleValue.intPart = (int16_t)val32;
      pReading->doubleValue.fracPart = 0;
      break;

    default:
      return CoreSensorUnexpectedError;
    }
    return CoreSensorSuccess;
  }

  return CoreSensorInvalidIndex;
}

boolean_t
admxrc6txReadXrmByte(
	Adb3CoreDeviceContext* pDevCtx,
  uint8_t address,
  uint8_t* pVal8)
{
	ModelRegsAdmxrc6tx* pModelRegs;
  DfMemoryHandle hModel;
  uint32_t cmd, val32 = 0;
  unsigned int n = 0;

  pModelRegs = pDevCtx->hardware.model.pAdmxrc6tx;
  hModel = pDevCtx->hardware.hModel;

  cmd = ADMXRC6Tx_V_I2CCTL_DISADDR(address) | ADMXRC6Tx_V_I2CCTL_IFADDR(0) | ADMXRC6Tx_V_I2CCTL_XRM | ADMXRC6Tx_V_I2CCTL_DISREAD;
  dfPciMemWrite32(hModel, &pModelRegs->i2cCtl, cmd);
  while (n < 2) {
    val32 = dfPciMemRead32(hModel, &pModelRegs->i2cCtl);
    if (val32 & ADMXRC6Tx_V_I2CCTL_READY) {
      break;
    }
    dfDelayThreadFor(dfMicrosecondsToTime(1000));
  }
  if (n == 2) {
    /* Timeout */
    return FALSE;
  }
  if (val32 & ADMXRC6Tx_V_I2CCTL_ERROR) {
    return FALSE;
  }

  *pVal8 = (uint8_t)val32;

  return TRUE;
}
