/*
** File: coreclock.c  
** Project: ADB3 core driver
** Purpose: Implements functions for clock generators in Core Interface.
**
** (C) Copyright Alpha Data 2009-2010
*/

#include <df.h>

#include "resource.h"
#include "coreclock.h"

/*
** Handle CoreClockGetFreq
*/

CoreClockStatus
coreClockGetFreq(
  void* pInterfaceContext,
  CoreTicket* pTicket,
  unsigned int clockIndex,
  uint64_t* pActualFrequency)
{
  Adb3CoreDeviceContext* pDevCtx = (Adb3CoreDeviceContext*)pInterfaceContext;

  if (clockIndex >= pDevCtx->info.bootstrap.numClockGen) {
    return CoreClockInvalidIndex;
  }

  if (!checkTicketValid(pDevCtx, pTicket, CORE_RESOURCE_CLOCKGEN, clockIndex)) {
    return CoreClockInvalidTicket;
  }

  if (0 != pDevCtx->clockGenerator[clockIndex].currentFrequency) {
    *pActualFrequency = pDevCtx->clockGenerator[clockIndex].currentFrequency;
  } else {
    return CoreClockHardwareError;
  }

  return CoreClockSuccess;
}

/*
** Handle CoreClockProgram
*/

typedef enum _State {
  StateIdle = 0,
  StatePreWait = 1,
  StateWaiting = 2,
  StatePostWait = 3,
  StateComplete = 4
} State;

/*
** This function executes "atomically", because the ICD2061 clock generator (that may be used on
** some models) must be programmed with a certain maximum period of time.
*/
static void*
clockProgramAtomic(
  DfInterruptObject* pInterruptObject, /* ignored */
  void* pArgument1,
  void* pArgument2)
{
  Adb3CoreDeviceContext* pDevCtx = (Adb3CoreDeviceContext*)pArgument1;
  ClockGeneratorContext* pDevClockCtx = (ClockGeneratorContext*)pArgument2;
  CoreClockStatus status;
  unsigned int clockIndex;
  uint32_t clockWord;

  clockIndex = pDevClockCtx->index;
  clockWord = pDevClockCtx->program.word;
  status = (CoreClockStatus)pDevCtx->methods.pClockProgram(pDevCtx, clockIndex, clockWord, pDevClockCtx->program.state == StatePostWait ? TRUE : FALSE);

  return (void*)(uintptr_t)status;
}

static void
clockProgramPostCallback(
  void* pTimerContext,
  void* pCallbackContext)
{
  Adb3CoreDeviceContext* pDevCtx = (Adb3CoreDeviceContext*)pTimerContext;
  ClockGeneratorContext* pDevClockCtx = (ClockGeneratorContext*)pCallbackContext;
  CoreClockStatus status;

  pDevClockCtx->program.state = StatePostWait;
  status = (CoreClockStatus)(uintptr_t)dfInterruptExecuteAtomic(&pDevCtx->interrupt.object, clockProgramAtomic, pDevCtx, pDevClockCtx);
  pDevClockCtx->program.state = StateComplete;
  pDevClockCtx->program.pCallback(pDevClockCtx->program.pCallbackTicket, pDevClockCtx->program.pCallbackContext, status);
}

CoreClockStatus
coreClockProgram(
  void* pInterfaceContext,
  CoreTicket* pTicket,
  unsigned int clockIndex,
  uint32_t clockWord,
  CoreClockProgCallback* pCallback,
  void* pContext)
{
  Adb3CoreDeviceContext* pDevCtx = (Adb3CoreDeviceContext*)pInterfaceContext;
  ClockGeneratorContext* pDevClockCtx;
  CoreClockStatus status = CoreClockSuccess;

  if (clockIndex >= pDevCtx->info.bootstrap.numClockGen) {
    return CoreClockInvalidIndex;
  }

  pDevClockCtx = &pDevCtx->clockGenerator[clockIndex];
  pDevClockCtx->program.state = StatePreWait;
  pDevClockCtx->program.pCallback = pCallback;
  pDevClockCtx->program.pCallbackTicket = pTicket;
  pDevClockCtx->program.pCallbackContext = pContext;
  pDevClockCtx->program.word = clockWord;
  status = (CoreClockStatus)(uintptr_t)dfInterruptExecuteAtomic(&pDevCtx->interrupt.object, clockProgramAtomic, pDevCtx, pDevClockCtx);
  if (CoreClockSuccess != status) {
    return status;
  }
  pDevClockCtx->program.state = StateWaiting;

  /* We allow 50 milliseconds before we check whether or not clock programming was successful */
  dfTimerSchedule(&pDevClockCtx->program.timer, dfTimeGet() + dfMicrosecondsToTime(50000), clockProgramPostCallback, pDevClockCtx);

  return status;
}

/*
** Handle CoreClockTestFreq
*/

CoreClockStatus
coreClockTestFreq(
  void* pInterfaceContext,
  unsigned int clockIndex,
  uint32_t flags,
  uint64_t frequencyReq,
  uint64_t frequencyMin,
  uint64_t frequencyMax,
  uint32_t* clockWord,
  uint64_t* pActualFrequency)
{
  Adb3CoreDeviceContext* pDevCtx = (Adb3CoreDeviceContext*)pInterfaceContext;

  if (clockIndex >= pDevCtx->info.bootstrap.numClockGen) {
    return CoreClockInvalidIndex;
  }

  return (CoreClockStatus)pDevCtx->methods.pClockWord(pDevCtx, clockIndex, flags, frequencyReq, frequencyMin, frequencyMax, clockWord, pActualFrequency);
}

unsigned int
coreGetNumClockGen(
  void* pInterfaceContext)
{
  Adb3CoreDeviceContext* pDevCtx = (Adb3CoreDeviceContext*)pInterfaceContext;

  return pDevCtx->info.bootstrap.numClockGen;
}
