/*
** File: flash.c  
** Project: ADB3 core driver
** Purpose: Implements functions for identifying, reading, writing and erasing the
**          Flash devices on a card.
**
** (C) Copyright Alpha Data 2009-2010
*/

#include "cfi.h"
#include "device.h"
#include "flash.h"
#include "flash_cfi.h"

static CoreFlashStatus
eraseBlock(
  Adb3CoreDeviceContext*  pDevCtx,
  unsigned int index,
  uint64_t blockAddress)
{
  CfiInformation* pInfo = pDevCtx->info.flash.pDetail;

  dfDebugPrint(3, ("eraseBlock: entered, pDevCtx=%p index=%lu blockAddress=0x%08lx_%08lx\n",
    (void*)pDevCtx, (unsigned long)index, dfSplitUint64(blockAddress)));

  switch (pInfo->validation.preferredAlgorithm) {
  case CFI_ALG_INTELSHARP_EXT:
    return flashCfiEraseBlock0001(pDevCtx, index, blockAddress);

  default:
    /* Don't understand this command set */
    return CoreFlashGeneralFailure;
  }
}

static CoreFlashStatus
readChunk(
  Adb3CoreDeviceContext*  pDevCtx,
  unsigned int index,
  uint64_t start,
  uint64_t length,
  void* pBuffer)
{
  uint8_t* p;
  uint32_t val32;
  uint8_t align;
  uint64_t startAligned;

  dfDebugPrint(5, ("readChunk: entered, index=%lu start=0x%08lx_%08lx length=0x%08lx_%08lx pBuffer=%p\n",
    (unsigned long)index, dfSplitUint64(start), dfSplitUint64(length), pBuffer));

  if (length == 0) {
    return CoreFlashSuccess;
  }

  p = (uint8_t*)pBuffer;
  switch (pDevCtx->info.bootstrap.flashWidth) {
  case 1U:
    /* Flash device has 8-bit data bus */
    startAligned = start;
    while (length) {
      pDevCtx->methods.pFlash(pDevCtx, index, FALSE, startAligned, &val32);
      *p++ = (uint8_t)val32;
      length--;
      startAligned++;
    }
    break;

  case 2U:
    /* Flash device has 16-bit data bus*/
    startAligned = start & ~0x1ULL;
    align = (uint8_t)start & 0x1U;
    /* Ensure flash address aligned to a 16-bit word */
    if (align) {
      pDevCtx->methods.pFlash(pDevCtx, index, FALSE, startAligned, &val32);
      *p++ = (uint8_t)(val32 >> 8U);
      length--;
      startAligned += 2;
    }
    /* Do bulk of 16-bit words */
    while (length >= 2) {
      /* Align 'start' to a 16-bit word */
      pDevCtx->methods.pFlash(pDevCtx, index, FALSE, startAligned, &val32);
      *p++ = (uint8_t)(val32 >> 0U);
      *p++ = (uint8_t)(val32 >> 8U);
      length -= 2;
      startAligned += 2;
    }
    /* Finish last byte */
    if (length & 0x1U) {
      pDevCtx->methods.pFlash(pDevCtx, index, FALSE, startAligned, &val32);
      *p++ = (uint8_t)(val32 >> 0U);
    }
    break;

  case 4U:
    /* Flash device has 32-bit data bus */
    startAligned = start & ~0x3ULL;
    align = (uint8_t)start & 0x3U;
    /* Ensure flash address aligned to a 32-bit word */
    switch (align) {
    case 0U:
      break;

    case 1U:
      pDevCtx->methods.pFlash(pDevCtx, index, FALSE, startAligned, &val32);
      *p++ = (uint8_t)(val32 >> 8U);
      *p++ = (uint8_t)(val32 >> 16U);
      *p++ = (uint8_t)(val32 >> 24U);
      length -= 3;
      startAligned += 4;
      break;

    case 2U:
      pDevCtx->methods.pFlash(pDevCtx, index, FALSE, startAligned, &val32);
      *p++ = (uint8_t)(val32 >> 16U);
      *p++ = (uint8_t)(val32 >> 24U);
      length -= 2;
      startAligned += 4;

    case 3U:
      pDevCtx->methods.pFlash(pDevCtx, index, FALSE, startAligned, &val32);
      *p++ = (uint8_t)(val32 >> 24U);
      length -= 1;
      startAligned += 4;
    }
    /* Do bulk of 32-bit words */
    while (length >= 4) {
      pDevCtx->methods.pFlash(pDevCtx, index, FALSE, startAligned, &val32);
      *p++ = (uint8_t)(val32 >> 0U);
      *p++ = (uint8_t)(val32 >> 8U);
      *p++ = (uint8_t)(val32 >> 16U);
      *p++ = (uint8_t)(val32 >> 24U);
      length -= 4;
      startAligned += 4;
    }
    /* Finish last few bytes */
    switch ((unsigned int)length & 0x3U) {
    case 0U:
      break;

    case 1U:
      pDevCtx->methods.pFlash(pDevCtx, index, FALSE, startAligned, &val32);
      *p++ = (uint8_t)(val32 >> 0U);
      break;

    case 2U:
      pDevCtx->methods.pFlash(pDevCtx, index, FALSE, startAligned, &val32);
      *p++ = (uint8_t)(val32 >> 0U);
      *p++ = (uint8_t)(val32 >> 8U);
      break;

    case 3U:
      pDevCtx->methods.pFlash(pDevCtx, index, FALSE, startAligned, &val32);
      *p++ = (uint8_t)(val32 >> 0U);
      *p++ = (uint8_t)(val32 >> 8U);
      *p++ = (uint8_t)(val32 >> 16U);
      break;
    }
    break;

  default:
    return CoreFlashGeneralFailure;
  }

  return CoreFlashSuccess;
}

static CoreFlashStatus
writeBlock(
  Adb3CoreDeviceContext*  pDevCtx,
  unsigned int index,
  uint64_t blockAddress,
  uint64_t blockSize,
  const uint8_t* pData)
{
  CfiInformation* pInfo = pDevCtx->info.flash.pDetail;

  dfDebugPrint(3, ("writeBlock: entered, index=%lu blockAddress=0x%08lx_%08lx blockSize=0x%08lx_%08lx pData=%p\n",
    (unsigned long)index, dfSplitUint64(blockAddress), dfSplitUint64(blockSize), pData));

  switch (pInfo->validation.preferredAlgorithm) {
  case CFI_ALG_INTELSHARP_EXT:
    return flashCfiWriteBlock0001(pDevCtx, index, blockAddress, blockSize, pData);

  default:
    /* Don't understand this command set */
    return CoreFlashGeneralFailure;
  }
}

/*
** -----------------------------------------------------------------
** Exported routines
** -----------------------------------------------------------------
*/

CoreFlashStatus
flashErase(
  Adb3CoreDeviceContext* pDevCtx,
  unsigned int index,
  uint64_t start,
  uint64_t length)
{
  CoreFlashStatus status = CoreFlashSuccess;
  uint64_t blockStart, blockSize, blockOffset;
  uint64_t i, chunk;
  uint8_t* pMergeBuffer = pDevCtx->flashControl.pMergeBuffer;

  dfDebugPrint(5, ("flashErase: index=%lu start=0x%08lx_%08lx length=0x%08lx_%08lx\n",
    (unsigned long)index, dfSplitUint64(start), dfSplitUint64(length)));

  if (index != 0) {
    return CoreFlashInvalidIndex;
  }
  if (!pDevCtx->info.flash.bPresent) {
    return CoreFlashNotPresent;
  }
  if (start + length < start) {
    /* Region wraps to 0 */
    return CoreFlashInvalidRegion;
  }
  if (start > pDevCtx->info.flash.totalSize || (start + length) > pDevCtx->info.flash.totalSize) {
    /* Region out of bounds of Flash chip */
    return CoreFlashInvalidRegion;
  }

  while (length) {
    /* Find the address and size of the current block */
    status = flashIdentifyBlock(pDevCtx, index, start, &blockStart, &blockSize);
    if (CoreFlashSuccess != status) {
      return status;
    }
    blockOffset = start - blockStart;
    chunk = blockSize - blockOffset;
    if (chunk > length) {
      chunk = length;
    }
    if (chunk != blockSize) {
      /* Less than entire block is to be erased - we must save and restore data */
      status = readChunk(pDevCtx, index, blockStart, blockSize, pMergeBuffer);
      if (status != CoreFlashSuccess) {
        goto done;
      }
      for (i = 0; i < chunk; i++) {
        pMergeBuffer[blockOffset + i] = 0xffU;
      }
    }
    /* Erase the block */
    status = eraseBlock(pDevCtx, index, blockStart);
    if (status != CoreFlashSuccess) {
      goto done;
    }
    if (chunk != blockSize) {
      /* Restore data to the block after block-erase */
      status = writeBlock(pDevCtx, index, blockStart, blockSize, pMergeBuffer);
      if (status != CoreFlashSuccess) {
        goto done;
      }
    }
    start += chunk;
    length -= chunk;
  }

done:
  return status;
}

CoreFlashStatus
flashIdentifyBlock(
  Adb3CoreDeviceContext* pDevCtx,
  unsigned int index,
  uint64_t location,
  uint64_t* pBlockStart,
  uint64_t* pBlockSize)
{
  CfiInformation* pInfo;
  uint8_t i;
  uint64_t startAddr, endAddr;
  uint64_t blockIndex;

  if (index != 0) {
    return CoreFlashInvalidIndex;
  }
  if (!pDevCtx->info.flash.bPresent) {
    return CoreFlashNotPresent;
  }
  if (location > pDevCtx->info.flash.totalSize) {
    /* Location out of bounds of Flash chip */
    return CoreFlashInvalidRegion;
  }

  pInfo = pDevCtx->info.flash.pDetail;
  while (NULL != pInfo) {
    if (location - pInfo->chipAddress >= pInfo->geometry.fixed.size) {
      /* The location is not in this die / device; advance to the next die */
      pInfo = pInfo->validation.pLink;
    } else {
      if (!pInfo->validation.bBlockEraseSupported) {
        /* Only full-chip erase is supported; treat entire chip as a single block */
        *pBlockStart = pInfo->chipAddress;
        *pBlockSize = pInfo->geometry.fixed.size;
        return CoreFlashSuccess;
      }

      if (pInfo->geometry.fixed.numEraseRegion == 0) {
        /* Should never get here, as zero erase regions => block-erase not supported */
        return CoreFlashGeneralFailure;
      }

      for (i = 0; i < pInfo->geometry.fixed.numEraseRegion; i++) {
        startAddr = pInfo->geometry.pEraseRegion[i].startAddress;
        endAddr = pInfo->geometry.pEraseRegion[i].startAddress + pInfo->geometry.pEraseRegion[i].blockSize * pInfo->geometry.pEraseRegion[i].numBlock;
        if (location >= startAddr && location < endAddr) {
          blockIndex = dfDivideUint64(location - startAddr, pInfo->geometry.pEraseRegion[i].blockSize);
          *pBlockStart = startAddr + blockIndex * pInfo->geometry.pEraseRegion[i].blockSize;
          *pBlockSize = pInfo->geometry.pEraseRegion[i].blockSize;
          return CoreFlashSuccess;
        }
      }

      /* If we get here, there must be an error in the data structure */
      return CoreFlashGeneralFailure;
    }
  }

  /* The requested location is outside the bounds of the Flash device */
  return CoreFlashInvalidRegion;
}

CoreFlashStatus
flashRead(
  Adb3CoreDeviceContext* pDevCtx,
  unsigned int index,
  uint64_t start,
  size_t length,
  void* pBuffer)
{
  dfDebugPrint(5, ("flashRead: index=%lu start=0x%lx length=%p pBuffer=%p\n",
    (unsigned long)index, (unsigned long)start, (void*)length, pBuffer));

  if (index != 0) {
    return CoreFlashInvalidIndex;
  }
  if (!pDevCtx->info.flash.bPresent) {
    return CoreFlashNotPresent;
  }
  if (start + length < start) {
    /* Region wraps to 0 */
    return CoreFlashInvalidRegion;
  }
  if (start > pDevCtx->info.flash.totalSize || (start + length) > pDevCtx->info.flash.totalSize) {
    /* Region out of bounds of Flash chip */
    return CoreFlashInvalidRegion;
  }

  return readChunk(pDevCtx, index, start, length, pBuffer);
}

CoreFlashStatus
flashWrite(
  Adb3CoreDeviceContext* pDevCtx,
  unsigned int index,
  uint64_t start,
  size_t length,
  const void* pData)
{
  CoreFlashStatus status = CoreFlashSuccess;
  uint64_t blockStart, blockSize, blockOffset;
  uint64_t i, chunk;
  uint8_t* p = (uint8_t*)pData;
  uint8_t* pMergeBuffer = pDevCtx->flashControl.pMergeBuffer;

  dfDebugPrint(5, ("flashWrite: index=%lu start=0x%08lx_%08lx length=0x%08lx_%08lx pData=%p\n",
    (unsigned long)index, dfSplitUint64(start), dfSplitUint64((uint64_t)length), pData));

  if (!pDevCtx->info.flash.bPresent) {
    return CoreFlashNotPresent;
  }

  if (index != 0) {
    return CoreFlashInvalidIndex;
  }
  if (start + length < start) {
    /* Region wraps to 0 */
    return CoreFlashInvalidRegion;
  }
  if (start > pDevCtx->info.flash.totalSize || (start + length) > pDevCtx->info.flash.totalSize) {
    /* Region out of bounds of Flash chip */
    return CoreFlashInvalidRegion;
  }

  while (length) {
    /* Find the address and size of the current block */
    status = flashIdentifyBlock(pDevCtx, index, start, &blockStart, &blockSize);
    if (CoreFlashSuccess != status) {
      return status;
    }
    blockOffset = start - blockStart;
    chunk = blockSize - blockOffset;
    if (chunk > length) {
      chunk = length;
    }
    if (chunk != blockSize) {
      /* Less than entire block is to be erased - we must save and restore data */
      status = readChunk(pDevCtx, index, blockStart, blockSize, pMergeBuffer);
      if (status != CoreFlashSuccess) {
        goto done;
      }
      /* Merge in new data */
      for (i = 0; i < chunk; i++) {
        pMergeBuffer[blockOffset + i] = *p++;
      }
    }
    /* Erase the block */
    status = eraseBlock(pDevCtx, index, blockStart);
    if (status != CoreFlashSuccess) {
      goto done;
    }
    /* Write the block with (possibly) merged data */
    status = writeBlock(pDevCtx, index, blockStart, blockSize, (chunk != blockSize) ? pMergeBuffer : pData);
    if (status != CoreFlashSuccess) {
      goto done;
    }
    start += chunk;
    length -= (size_t)chunk; /* Cast is safe since we forced chunk < length above */
  }

done:
  return status;
}
