/*
** File: identify.c  
** Project: ADB3 core driver
** Purpose: Identify device given Hardware ID, and set the bootstrap method of
**          the device context appropriately for that Hardware ID.
**
** (C) Copyright Alpha Data 2009-2010
*/

#include <df.h>

#include "device.h"
#include "device_defs.h"
#include "identify.h"
#include "model_boot.h"

#if !defined(ADB3_CUSTOM_ONLY)
static boolean_t
identifyPci9656(
	Adb3CoreDeviceContext* pDevCtx,
	DfHardwareId* pId,
  boolean_t* pbMatch)
{
	dfDebugPrint(1, ("identifyPci9656: entered\n"));

	switch (pId->variant.pci.subDevice) {
	case PCI_SUBDEVICE_PCI9656_ADMXRC2:
		dfDebugPrint(0, ("identifyPci9656: identified ADM-XRC-II\n"));
		pDevCtx->methods.pModelBootstrap = adb3CoreBootstrapAdmxrc2;
    *pbMatch = TRUE;
		return TRUE;

  default:
		dfDebugPrint(0, ("*** identifyPci9656: unknown PCI9656 device\n"));
		return FALSE;
	}
}

static boolean_t
identifyPlxTech(
	Adb3CoreDeviceContext* pDevCtx,
	DfHardwareId* pId,
  boolean_t* pbMatch)
{
	dfDebugPrint(1, ("identifyAlphaData: entered\n"));

	switch (pId->variant.pci.device) {
	case PCI_DEVICE_PLXTECH_PCI9656:
		dfDebugPrint(1, ("identifyAlphaData: identified PCI9656\n"));
    return identifyPci9656(pDevCtx, pId, pbMatch);

  default:
		break;
	}

	dfDebugPrint(0, ("***  identifyAlphaData: failed to recognize device\n"));

	return FALSE;
}

static boolean_t
identifyAdb3(
	Adb3CoreDeviceContext* pDevCtx,
	DfHardwareId* pId,
  boolean_t* pbMatch)
{
	dfDebugPrint(1, ("identifyAdb3: entered\n"));

	switch (pId->variant.pci.subDevice) {
	case PCI_SUBDEVICE_ADB3_ADPEXRC5T:
		dfDebugPrint(0, ("identifyAdb3: identified ADPe-XRC-5T\n"));
		pDevCtx->methods.pModelBootstrap = adb3CoreBootstrapAdpexrc5t;
    *pbMatch = TRUE;
		return TRUE;

	case PCI_SUBDEVICE_ADB3_ADMXRC6TL_LX50T:
		dfDebugPrint(0, ("identifyAdb3: identified ADM-XRC-6TL (5VLX50T bridge)\n"));
		pDevCtx->methods.pModelBootstrap = adb3CoreBootstrapAdmxrc6tl;
    *pbMatch = TRUE;
		return TRUE;

  case PCI_SUBDEVICE_ADB3_ADMXRC6TL_FX50T:
		dfDebugPrint(0, ("identifyAdb3: identified ADM-XRC-6TL (5VFX70T bridge)\n"));
		pDevCtx->methods.pModelBootstrap = adb3CoreBootstrapAdmxrc6tl;
    *pbMatch = TRUE;
		return TRUE;

  case PCI_SUBDEVICE_ADB3_ADMXRC6T1:
		dfDebugPrint(0, ("identifyAdb3: identified ADM-XRC-6T1\n"));
		pDevCtx->methods.pModelBootstrap = adb3CoreBootstrapAdmxrc6t1;
    *pbMatch = TRUE;
		return TRUE;

  case PCI_SUBDEVICE_ADB3_ADMXRC6T1_BERT:
		dfDebugPrint(0, ("identifyAdb3: identified ADM-XRC-6T1 (BERT configuration)\n"));
		pDevCtx->methods.pModelBootstrap = adb3CoreBootstrapAdmxrc6t1;
    *pbMatch = TRUE;
		return TRUE;

  default:
		dfDebugPrint(0, ("identifyAdb3: identified generic ADB3 device\n"));
		pDevCtx->methods.pModelBootstrap = adb3CoreBootstrapGenericAdb3;
    *pbMatch = TRUE;
		return TRUE;
	}
}

static boolean_t
identifyAlphaData(
	Adb3CoreDeviceContext* pDevCtx,
	DfHardwareId* pId,
  boolean_t* pbMatch)
{
	dfDebugPrint(1, ("identifyAlphaData: entered\n"));

	switch (pId->variant.pci.device) {
	case PCI_DEVICE_ADATA_ADB3:
		dfDebugPrint(1, ("identifyAlphaData: identified ADB3\n"));
    return identifyAdb3(pDevCtx, pId, pbMatch);

	default:
		break;
	}

	dfDebugPrint(0, ("***  identifyAlphaData: failed to recognize device\n"));

	return FALSE;
}
#endif

boolean_t
identifyDevice(
	Adb3CoreDeviceContext* pDevCtx,
	DfHardwareId* pId,
  boolean_t* pbMatch)
{
	dfDebugPrint(1, ("identifyDevice: entered\n"));

#if defined(ADB3_CUSTOM)
  if (identifyDeviceCustom(pDevCtx, pId, pbMatch)) {
    return TRUE;
  }
#endif

#if !defined(ADB3_CUSTOM_ONLY)
	if (pId->busType == DfBusPci) {
    switch (pId->variant.pci.vendor) {
    case PCI_VENDOR_ADATA:
    	dfDebugPrint(1, ("identifyDevice: vendor is ALPHA DATA\n"));
			identifyAlphaData(pDevCtx, pId, pbMatch);
      break;

    case PCI_VENDOR_PLXTECH:
    	dfDebugPrint(1, ("identifyDevice: vendor is PLX Tech.\n"));
			identifyPlxTech(pDevCtx, pId, pbMatch);
      break;

    default:
    	/* Vendor not recognized */
      return FALSE;
    }
	}
#endif

  return FALSE;
}
