/*
** File: model_admxrc2.c  
** Project: ADB3 core driver
** Purpose: ADM-XRC-II-specific code, including function to bootstrap device context.
**
** (C) Copyright Alpha Data 2009-2010
*/

#include <df.h>

#include "admxrc2.h"
#include "coredma.h"
#include "device.h"
#include "flash.h"
#include "flash_cfi.h"
#include "icd2061.h"
#include "model_boot.h"
#include "pci9xxx.h"
#include "pci9xxx_93x6.h"

#define VPD_OFFSET (0x6AU)

static CoreVpdStatus
readWriteVpd(
  Adb3CoreDeviceContext* pDevCtx,
  boolean_t bWrite,
  size_t address,
  size_t length,
  void* pData);

static void
dumpDmaDescriptors(
  Adb3CoreDeviceContext* pDevCtx,
  Pci9xxxDmaDescriptor* pDescriptors,
  uint64_t descBus,
  unsigned int count)
{
  Pci9xxxDmaDescriptor* pCurrent;
  unsigned int i;

  if (sizeof(void*) == 4) {
    dfDebugPrint(0, ("  ### DescVirt   DescBus             [ PA       LA       SIZ      NEXT     ]\n"));
    for (i = 0; i < count; i++) {
      pCurrent = &pDescriptors[i];
      dfDebugPrint(0, ("  %03lu 0x%8p 0x%08lx_%08lx   %08lx %08lx %08lx %08lx\n",
        (unsigned long)i, (void*)pCurrent, dfSplitUint64(descBus),
        (unsigned long)dfLe32ToCpu(pCurrent->pci), (unsigned long)dfLe32ToCpu(pCurrent->local),
        (unsigned long)dfLe32ToCpu(pCurrent->count), (unsigned long)dfLe32ToCpu(pCurrent->link)));
      descBus += sizeof(Pci9xxxDmaDescriptor);
    }
  } else {
    dfDebugPrint(0, ("  ### DescVirt           DescBus             [ PA       LA       SIZ      NEXT     ]\n"));
    for (i = 0; i < count; i++) {
      pCurrent = &pDescriptors[i];
      dfDebugPrint(0, ("  %03lu 0x%16p 0x%08lx_%08lx   %08lx %08lx %08lx %08lx\n",
        (unsigned long)i, (void*)pCurrent, dfSplitUint64(descBus),
        (unsigned long)dfLe32ToCpu(pCurrent->pci), (unsigned long)dfLe32ToCpu(pCurrent->local),
        (unsigned long)dfLe32ToCpu(pCurrent->count), (unsigned long)dfLe32ToCpu(pCurrent->link)));
      descBus += sizeof(Pci9xxxDmaDescriptor);
    }
  }
}

static void
dumpDmaRegisters(
  Adb3CoreDeviceContext* pDevCtx,
  unsigned int channel)
{
  Pci9xxxDmaRegisters* pDmaRegs = &pDevCtx->hardware.bridge.pPci9656->DMA[channel];
  uint32_t mode, padr, ladr, size, dpr;
  uint16_t dmacsr;

  mode = pDmaRegs->MODE;
  padr = pDmaRegs->PADR;
  ladr = pDmaRegs->LADR;
  size = pDmaRegs->SIZE;
  dpr = pDmaRegs->DPR;
  dmacsr = pDevCtx->hardware.bridge.pPci9656->DMACSR;
  dfDebugPrint(0, ("  channel=%u MODE=%08lx PADR=%08lx LADR=%08lx SIZE=%08lx DPR=%08lx DMACSR=0x%04lx\n",
    channel, (unsigned long)mode, (unsigned long)padr, (unsigned long)ladr, (unsigned long)size, (unsigned long)dpr, (unsigned long)dmacsr));
}

ClockProgramStatus
clockProgram(
  Adb3CoreDeviceContext* pDevCtx,
  unsigned int clockIndex,
  uint32_t clockWord,
  boolean_t bAfter)
{
  ModelRegsAdmxrc2* pModelRegs;
  DfMemoryHandle hModel;
  unsigned int i;
  uint8_t val8;
		
  dfDebugPrint(3, ("clockProgram: entered, clockIndex=%lu clockWord=0x%lx bAfter=%s\n",
    (unsigned long)clockIndex, (unsigned long)clockWord, bAfter ? "TRUE" : "FALSE"));

  if (clockIndex >= 2) {
    return ClockProgramInvalidIndex;
  }

  pModelRegs = pDevCtx->hardware.model.pAdmxrc2;
  hModel = pDevCtx->hardware.hModel;

  if (bAfter) {
    if (dfPciMemRead8(hModel, &pModelRegs->ccon) & 0x10) {
      return ClockProgramHardwareError;
    }
  } else {
    dfPciMemWrite8(hModel, &pModelRegs->ccon, ICD2061_CLK | ICD2061_DATA);
    (void)dfPciMemRead8(hModel, &pModelRegs->ccon);
    
    /*
    ** Send unlock sequence to ICD2061; at least five 0->1 CLK transitions with DATA
    **  held at 1, and than a 0->1 CLK transition with DATA at 0.
    */
    for (i = 0; i < 6; i++) {
      dfPciMemWrite8(hModel, &pModelRegs->ccon, ICD2061_DATA);
      (void)dfPciMemRead8(hModel, &pModelRegs->ccon);
      dfPciMemWrite8(hModel, &pModelRegs->ccon, ICD2061_CLK | ICD2061_DATA);
      (void)dfPciMemRead8(hModel, &pModelRegs->ccon);
    }
    dfPciMemWrite8(hModel, &pModelRegs->ccon, ICD2061_DATA);
    (void)dfPciMemRead8(hModel, &pModelRegs->ccon);
    dfPciMemWrite8(hModel, &pModelRegs->ccon, 0);
    (void)dfPciMemRead8(hModel, &pModelRegs->ccon);
    dfPciMemWrite8(hModel, &pModelRegs->ccon, ICD2061_CLK);
    (void)dfPciMemRead8(hModel, &pModelRegs->ccon);
    dfPciMemWrite8(hModel, &pModelRegs->ccon, 0);
    (void)dfPciMemRead8(hModel, &pModelRegs->ccon);
    dfPciMemWrite8(hModel, &pModelRegs->ccon, ICD2061_CLK);
    (void)dfPciMemRead8(hModel, &pModelRegs->ccon);
    
    /* Send the data bits to ICD2061 (uses a modified Manchester encoding) */
    for (i = 0; i < 24; i++) {
      if (clockWord & 0x1U) {
        val8 = 0;
      } else {
        val8 = 2;
      }
      clockWord >>= 1;
      
      dfPciMemWrite8(hModel, &pModelRegs->ccon, val8 | ICD2061_CLK);
      (void)dfPciMemRead8(hModel, &pModelRegs->ccon);
      dfPciMemWrite8(hModel, &pModelRegs->ccon, val8);
      (void)dfPciMemRead8(hModel, &pModelRegs->ccon);
      dfPciMemWrite8(hModel, &pModelRegs->ccon, val8 ^ ICD2061_DATA);
      (void)dfPciMemRead8(hModel, &pModelRegs->ccon);
      dfPciMemWrite8(hModel, &pModelRegs->ccon, (val8 ^ ICD2061_DATA) | ICD2061_CLK);
      (void)dfPciMemRead8(hModel, &pModelRegs->ccon);
    }
    
    /*
    ** Send stop bit to ICD2061; this is a 1->0 and 0->1 transition of CLK while DATA
    ** is 1.
    */
    dfPciMemWrite8(hModel, &pModelRegs->ccon, ICD2061_CLK | ICD2061_DATA);
    (void)dfPciMemRead8(hModel, &pModelRegs->ccon);
    dfPciMemWrite8(hModel, &pModelRegs->ccon, ICD2061_DATA);
    (void)dfPciMemRead8(hModel, &pModelRegs->ccon);
    dfPciMemWrite8(hModel, &pModelRegs->ccon, ICD2061_CLK | ICD2061_DATA);
    (void)dfPciMemRead8(hModel, &pModelRegs->ccon);
    dfPciMemWrite8(hModel, &pModelRegs->ccon, 0);
    (void)dfPciMemRead8(hModel, &pModelRegs->ccon);
  }
  
  return ClockProgramSuccess;
}

static ClockWordStatus
clockWord(
  Adb3CoreDeviceContext* pDevCtx,
  unsigned int clockIndex,
  uint32_t flags,
  uint64_t frequencyReq,
  uint64_t frequencyMin,
  uint64_t frequencyMax,
  uint32_t* pClockWord,
  uint64_t* pActualFrequency)
{
  uint32_t refFrequency;
  VpdAdmxrc2Rev0* pVpd = &pDevCtx->info.vpd.pData->admxrc2.rev0;
  
  if (pVpd->refClock == 0) {
    refFrequency = 16666667U;
  } else {
    refFrequency = 14318181U;
  }
  return clockWordIcd2061(pDevCtx, clockIndex, flags, refFrequency, frequencyReq, frequencyMin, frequencyMax, pClockWord, pActualFrequency);
}

static void
dmaList(
  Adb3CoreDeviceContext* pDevCtx,
  boolean_t bWriteToDevice,
  boolean_t bFixedLocal,
  DfDmaMapperNode* pTable,
  unsigned int tableLength,
  DmaNodeTablePosition* pTablePosition,
  void* pDescriptorBuffer,
  unsigned int maxDescriptor,
  uint64_t descriptorBusAddress)
{
  Pci9xxxDmaDescriptor* pDescriptor = (Pci9xxxDmaDescriptor*)pDescriptorBuffer;
  DfDmaMapperNode* pTableTmp;
  uint64_t busAddress, nextBusAddress, tableBusAddress, localAddress = 0U;
  unsigned int i;
  uint32_t flags;
  unsigned int numDesc = 0;
  unsigned int index;
  size_t offset, remaining;

  dfDebugPrint(7, ("dmaList: entered, pDevCtx=%p bWriteToDevice=%s bFixedLocal=%s pTable=%p tableLength=%u *pTablePosition={%u, 0x%llx} pDescriptorBuffer=%p maxDescriptor=%u descriptorBusAddrss=0x%llx\n",
    (void*)pDevCtx, bWriteToDevice ? "TRUE" : "FALSE", bFixedLocal ? "TRUE" : "FALSE", (void*)pTable, tableLength, pTablePosition->index, (unsigned long long)pTablePosition->offset, pDescriptorBuffer, maxDescriptor, (unsigned long long)descriptorBusAddress));

  flags = PCI9xxxDESCF_FETCH_PCI | (bWriteToDevice ? PCI9xxxDESCF_PCI_TO_LOCAL : PCI9xxxDESCF_LOCAL_TO_PCI);
  busAddress = descriptorBusAddress;
  index = pTablePosition->index;
  offset = pTablePosition->offset;
  pTableTmp = pTable + pTablePosition->index;
  for (i = index; i < tableLength && numDesc < maxDescriptor; i++, numDesc++) {
    nextBusAddress = busAddress + sizeof(Pci9xxxDmaDescriptor);
    tableBusAddress = pTableTmp->busAddress + offset;
    localAddress = pTableTmp->localAddress;
    if (!bFixedLocal) {
      localAddress += offset;
    }
    remaining = pTableTmp->length - offset;
    pDescriptor->pci = dfCpuToLe32((uint32_t)tableBusAddress);
    pDescriptor->local = dfCpuToLe32((uint32_t)localAddress);
    pDescriptor->count = dfCpuToLe32((uint32_t)remaining);
    pDescriptor->link = dfCpuToLe32((uint32_t)nextBusAddress | flags);
    pTableTmp++;
    busAddress = nextBusAddress;
    offset = 0;
    pDescriptor++;
  }
  pDescriptor--;
  pDescriptor->link = dfCpuToLe32(flags | PCI9xxxDESCF_ENDCHAIN);

  /* Cast is safe as pTableTmp cannot be much greater than pTable */
  pTablePosition->index = (unsigned int)(pTableTmp - pTable);
  pTablePosition->offset = offset;

#if DF_DBG_BUILD
  if (g_dfDebugLevel >= 7) {
    dfDebugPrint(0, ("dmaList: dump of DMA engine linked list\n"));
    dumpDmaDescriptors(pDevCtx, (Pci9xxxDmaDescriptor*)pDescriptorBuffer, descriptorBusAddress, numDesc);
  }
#endif
}

static void
dmaTransfer(
  Adb3CoreDeviceContext* pDevCtx,
  unsigned int channel,
  boolean_t bStart,
  void* pDescriptorBuffer,       /* Valid iff 'bStart' is TRUE */
  uint64_t descriptorBusAddress) /* Valid iff 'bStart' is TRUE */
{
  const unsigned int maxAbortPoll = 20;
  Pci9656Registers* pRegs;
  Pci9xxxDmaRegisters* pDmaRegs;
  DfMemoryHandle hBridge;
  uint32_t mode;
  uint16_t dmacsr, mask, val16;
  unsigned int i;

  dfDebugPrint(7, ("dmaTransfer: entered, pDevCtx=%p channel=%lu bStart=%s pDescriptorBuffer=%p busAddress=0x%08lx_%08lx\n",
    (void*)pDevCtx, (unsigned long)channel, bStart ? "TRUE" : "FALSE", pDescriptorBuffer, dfSplitUint64(descriptorBusAddress)));

  pRegs = pDevCtx->hardware.bridge.pPci9656;
  hBridge = pDevCtx->hardware.hBridge;
  pDmaRegs = &pRegs->DMA[channel];

  if (bStart) {
    mode = PCI9656_DMAMODE_32BIT | PCI9656_DMAMODE_USE_READY | PCI9656_DMAMODE_USE_BTERM | PCI9656_DMAMODE_ENABLE_BURST |
      PCI9656_DMAMODE_CHAINED | PCI9656_DMAMODE_DONE_INTERRUPT | PCI9656_DMAMODE_DEMAND | PCI9656_DMAMODE_WRITE_INVALIDATE |
      PCI9656_DMAMODE_INTERRUPT_PCI;
    if (pDevCtx->dma.channel[channel].curr.parameters.bFixedLocalAddress) {
      mode |= PCI9656_DMAMODE_FIXED_LOCAL;
    }
    /* Set current transfer mode */
    dfPciMemWrite32(hBridge, &pDmaRegs->MODE, mode);
    /* Set current transfer size to 0 */
    dfPciMemWrite32(hBridge, &pDmaRegs->SIZE, 0);
    /* Set the descriptor pointer & flags */
    dfPciMemWrite32(hBridge, &pDmaRegs->DPR, (uint32_t)descriptorBusAddress | PCI9xxxDESCF_FETCH_PCI);
    dfPciMemRead32(hBridge, &pDmaRegs->DPR);
    
#if DF_DBG_BUILD
    if (g_dfDebugLevel >= 7) {
      dfDebugPrint(0, ("dmaTransfer: DMA channel %lu registers before starting on hardware:\n", (unsigned long)channel));
      dumpDmaRegisters(pDevCtx, channel);
    }
#endif

    /* Start the DMA transfer */
    mask = channel ? 0x0200U : 0002U;
    dfPciMemWrite16(hBridge, &pRegs->DMACSR, mask | pDevCtx->hardware.shadow.bridge.pci9656.dmacsr);
    (void)dfPciMemRead16(hBridge, &pRegs->DMACSR); 

#if DF_DBG_BUILD
    if (g_dfDebugLevel >= 7) {
      dfDebugPrint(0, ("dmaTransfer: DMA channel %lu registers after starting on hardware:\n", (unsigned long)channel));
      dumpDmaRegisters(pDevCtx, channel);
    }
#endif
  } else {
#if DF_DBG_BUILD
    if (g_dfDebugLevel >= 7) {
      dfDebugPrint(0, ("dmaTransfer: DMA channel %lu registers before attempt to abort:\n", (unsigned long)channel));
      dumpDmaRegisters(pDevCtx, channel);
    }
#endif

    mask = channel ? 0x0100U : 0001U;
    /* First, disable the DMA channel */
    dmacsr = pDevCtx->hardware.shadow.bridge.pci9656.dmacsr & ~mask;
    dfPciMemWrite16(hBridge, &pRegs->DMACSR, dmacsr);
    /* Now check if DMA channel is active or not */
    val16 = dfPciMemRead16(hBridge, &pRegs->DMACSR);
    if (val16 & (mask << 4)) {
      /* DMA is not active so must have already finished */
      dfDebugPrint(1, ("dmaTransfer: nothing to abort, channel=%lu\n", (unsigned long)channel));
    } else {
      /* DMA is active - abort it */
      dfPciMemWrite16(hBridge, &pRegs->DMACSR, dmacsr | (mask << 2));
      /* Poll for PCI9080/PCI9656 aborting the DMA */
      for (i = 0; i < maxAbortPoll; i++) {
        val16 = dfPciMemRead16(hBridge, &pRegs->DMACSR);
        if (val16 & (mask << 4)) {
          dfDebugPrint(2, ("dmaTransfer: aborted, channel=%lu\n", (unsigned long)channel));
          break;
        }
      }
      if (maxAbortPoll == i) {
        /* No! Not much we can do, though */
        dfDebugPrint(0, ("*** dmaTransfer: abort failed, channel=%lu\n", (unsigned long)channel));
      }
    }
    /* Reenable the DMA channel */
    dmacsr |= mask;
    dfPciMemWrite16(hBridge, &pRegs->DMACSR, dmacsr);
    
#if DF_DBG_BUILD
    if (g_dfDebugLevel >= 7) {
      dfDebugPrint(0, ("dmaTransfer: DMA channel %lu registers after attempt to abort:\n", (unsigned long)channel));
      dumpDmaRegisters(pDevCtx, channel);
    }
#endif
  }
}

static void
enableInterrupts(
  Adb3CoreDeviceContext* pDevCtx,
  boolean_t bEnable)
{
  Pci9656Registers* pRegs;
  ModelRegsAdmxrc2* pModelRegs;
  DfMemoryHandle hBridge, hModel;
  uint32_t intcsr;
  uint16_t dmacsr;
  uint8_t imask;
  
  pRegs = pDevCtx->hardware.bridge.pPci9656;
  hBridge = pDevCtx->hardware.hBridge;
  pModelRegs = pDevCtx->hardware.model.pAdmxrc2;
  hModel = pDevCtx->hardware.hModel;
  dfDebugPrint(1, ("enableInterrupts: entered bEnable=%s\n", bEnable ? "TRUE" : "FALSE"));
  
  if (bEnable) {
    pDevCtx->hardware.shadow.model.admxrc2.ienab = 0x1U;
    dfPciMemWrite8(hBridge, &pModelRegs->imset, 0xffU);
    dfPciMemWrite8(hBridge, &pModelRegs->imclr, pDevCtx->hardware.shadow.model.admxrc2.ienab);
    dfPciMemWrite8(hBridge, &pModelRegs->icon, 0xffU);
    imask = dfPciMemRead8(hBridge, &pModelRegs->imset);
    dfDebugPrint(2, ("enableInterrupts: CPLD interrupts enabled, IMASK=0x%02lx\n", (unsigned long)imask));

    pDevCtx->hardware.shadow.bridge.pci9656.intcsr = PLX_INTCSR_PCI_IE | PLX_INTCSR_LOC_IE | PLX_INTCSR_DMA0_IE | PLX_INTCSR_DMA1_IE;
    pDevCtx->hardware.shadow.bridge.pci9656.intcsrMask = PLX_INTCSR_LOC_INT | PLX_INTCSR_DMA0_INT | PLX_INTCSR_DMA1_INT;
    dfPciMemWrite32(hBridge, &pRegs->INTCSR, pDevCtx->hardware.shadow.bridge.pci9656.intcsr);
    intcsr = dfPciMemRead32(hBridge, &pRegs->INTCSR);
    dfDebugPrint(2, ("enableInterrupts: interrupts enabled, INTCSR=0x%08lx\n", (unsigned long)intcsr));
    
    pDevCtx->hardware.shadow.bridge.pci9656.dmacsr = 0x0101U;
    dfPciMemWrite16(hBridge, &pRegs->DMACSR, pDevCtx->hardware.shadow.bridge.pci9656.dmacsr);
    dmacsr = dfPciMemRead16(hBridge, &pRegs->DMACSR);
    dfDebugPrint(2, ("enableInterrupts: DMA engines enabled, DMACSR=0x%04lx\n", (unsigned long)dmacsr));
  } else {
    pDevCtx->hardware.shadow.bridge.pci9656.dmacsr = 0U;
    dfPciMemWrite16(hBridge, &pRegs->DMACSR, pDevCtx->hardware.shadow.bridge.pci9656.dmacsr);
    dmacsr = dfPciMemRead16(hBridge, &pRegs->DMACSR);
    dfDebugPrint(2, ("enableInterrupts: DMA engines disabled, DMACSR=0x%04lx\n", (unsigned long)dmacsr));

    pDevCtx->hardware.shadow.bridge.pci9656.intcsr = 0U;
    pDevCtx->hardware.shadow.bridge.pci9656.intcsrMask = 0U;
    dfPciMemWrite32(hBridge, &pRegs->INTCSR, pDevCtx->hardware.shadow.bridge.pci9656.intcsr);
    intcsr = dfPciMemRead32(hBridge, &pRegs->INTCSR);
    dfDebugPrint(2, ("enableInterrupts: interrupts disabled, INTCSR=0x%08lx\n", (unsigned long)intcsr));
    
    dfPciMemWrite8(hBridge, &pModelRegs->imset, 0xffU);
    imask = dfPciMemRead8(hBridge, &pModelRegs->imset);
    dfDebugPrint(2, ("enableInterrupts: CPLD interrupts disabled, IMASK=0x%02lx\n", (unsigned long)imask));
  }
}

static void
flash(
  Adb3CoreDeviceContext* pDevCtx,
  unsigned int bankIndex,
  boolean_t bWrite,
  uint64_t location,
  uint32_t* pValue)
{
  ModelRegsAdmxrc2* pModelRegs;
  DfMemoryHandle hModel;
  uint8_t page = (uint8_t)(location >> 21);
  uint8_t* pFlash;

  dfDebugPrint(9, ("flash: bankIndex=%lu bWrite=%lu location=0x%08lx_%08lx *pValue=0x%lx\n",
    (unsigned long)bankIndex, (unsigned long)bWrite, dfSplitUint64(location), (unsigned long)*pValue));

  if (bankIndex != 0 || location >= 0x10000000U) {
    return;
  }

  pModelRegs = pDevCtx->hardware.model.pAdmxrc2;
  hModel = pDevCtx->hardware.hModel;
  pFlash = ((uint8_t*)pDevCtx->hardware.model.pAdmxrc2) + 0x200000U;

  /* Set Flash page register */
  dfPciMemWrite8(hModel, &pModelRegs->flpage, page & 0x7fU);
  (void)dfPciMemRead8(hModel, &pModelRegs->flpage);

  location &= 0x1fffffU;
  if (bWrite) {
    dfPciMemWrite8(hModel, pFlash + location, (uint8_t)*pValue);
    (void)dfPciMemRead8(hModel, &pModelRegs->flpage);
  } else {
    *pValue = (uint32_t)dfPciMemRead8(hModel, pFlash + location);
  }
}

FpgaControlStatus
fpgaControl(
  Adb3CoreDeviceContext* pDevCtx,
  unsigned int index,
  CoreFpgaControlOp opCode,
  boolean_t* pbValue)
{
  ModelRegsAdmxrc2* pModelRegs;
  DfMemoryHandle hModel;
  uint8_t val8;

  dfDebugPrint(9, ("fpgaControl: index=%lu opCode=%lu pbValue=%p\n",
    (unsigned long)index, (unsigned long)opCode, pbValue));
  
  if (0 != index) {
    return FpgaControlInvalidIndex;
  }

  pModelRegs = pDevCtx->hardware.model.pAdmxrc2;
  hModel = pDevCtx->hardware.hModel;

  switch (opCode) {
  case CoreFpgaControlOpGetDone:
    val8 = dfPciMemRead8(hModel, &pModelRegs->fcon);
    *pbValue = (val8 & ADMXRC2_FCON_DONE) ? TRUE : FALSE;
    dfDebugPrint(9, ("fpgaControl: opCode=GetDone val8=0x%lx *pbValue=%s\n", (unsigned long)val8, *pbValue ? "TRUE" : "FALSE"));
    break;

  case CoreFpgaControlOpGetLinkStatus:
    *pbValue = TRUE;
    dfDebugPrint(9, ("fpgaControl: opCode=GetLinkStat *pbValue=%s\n", *pbValue ? "TRUE" : "FALSE"));
    break;

  case CoreFpgaControlOpGetInit:
    val8 = dfPciMemRead8(hModel, &pModelRegs->fcon);
    *pbValue = (val8 & ADMXRC2_FCON_INIT) ? TRUE : FALSE;
    dfDebugPrint(9, ("fpgaControl: opCode=GetInit val8=0x%lx *pbValue=%s\n", (unsigned long)val8, *pbValue ? "TRUE" : "FALSE"));
    break;

  case CoreFpgaControlOpSetProg:
    val8 = pDevCtx->hardware.shadow.model.admxrc2.fcon;
    if (*pbValue) {
      val8 = val8 | ADMXRC2_FCON_PROG;
    } else {
      val8 = val8 & ~ADMXRC2_FCON_PROG;
    }
    dfDebugPrint(9, ("fpgaControlCommonPlxBased: opCode=SetProg val8=0x%lx *pbValue=%s\n", (unsigned long)val8, *pbValue ? "TRUE" : "FALSE"));
    dfPciMemWrite8(hModel, &pModelRegs->fcon, val8);
    (void)dfPciMemRead8(hModel, &pModelRegs->fcon);
    break;

  default:
    return FpgaControlInvalidOp;
  }

  return FpgaControlSuccess;
}

FpgaSelectMapStatus
fpgaSelectMap(
  Adb3CoreDeviceContext* pDevCtx,
  unsigned int index,
  boolean_t bWrite,
  size_t length,
  void* pBuffer)
{
  const size_t mask32 = 0x3U;
  ModelRegsAdmxrc2* pModelRegs;
  DfMemoryHandle hModel;
  uint8_t* pBuffer8;
  uint32_t* pBuffer32;
  size_t length32, align32;
  uint8_t val8;
  uint32_t val32;

  dfDebugPrint(9, ("fpgaSelectMap: index=%lu bWrite=%s length=%llu(0x%08lx_%08lx) pBuffer=%p\n",
    (unsigned long)index, bWrite ? "TRUE" : "FALSE", (unsigned long long)length, dfSplitUint64((uint64_t)length), pBuffer));
  
  if (0 != index) {
    return FpgaSelectMapInvalidIndex;
  }

  pModelRegs = pDevCtx->hardware.model.pAdmxrc2;
  hModel = pDevCtx->hardware.hModel;
  pBuffer8 = (uint8_t*)pBuffer;

  if (bWrite) {
    /* Align pointer to 32-bit boundary using 8-bit transfers */
    align32 = (size_t)(uintptr_t)pBuffer & mask32;
    length -= align32;
    while (align32--) {
      val8 = *pBuffer8++;
      dfPciMemWrite8(hModel, &pModelRegs->smap[0], val8);
    }
    
    /* Do bulk of data using 32-bit transfers */
    length32 = length & ~mask32;
    pBuffer32 = (uint32_t*)pBuffer8;
    length -= length32;
    while (length32) {
      val32 = dfLe32ToCpu(*pBuffer32++);
      dfPciMemWrite32(hModel, (uint32_t*)&pModelRegs->smap[0], val32);
      length32 -= 4;
    }
    
    /* Do remainder using 8-bit transfers */
    pBuffer8 = (uint8_t*)pBuffer32;
    while (length--) {
      val8 = *pBuffer8++;
      dfPciMemWrite8(hModel, &pModelRegs->smap[0], val8);
    }
    
    (void)dfPciMemRead8(hModel, &pModelRegs->fcon); /* Ensure above writes have propagated to destination */
  } else {
    while (length--) {
      val8 = dfPciMemRead8(hModel, &pModelRegs->smap[0]);
      *pBuffer8++ = val8;
    }
  }
  
  return FpgaSelectMapSuccess;
}

static boolean_t
initHardware(
  Adb3CoreDeviceContext* pDevCtx,
  boolean_t bInit)
{
  CoreVpdStatus status;
  Pci9656Registers* pRegs;
  DfMemoryHandle hBridge;
  uint32_t val32;
  uint32_t ls0b, ls0r, ls1b, ls1r;
  VpdAdmxrc2Rev0* pVpd, vpdRaw;

  dfDebugPrint(1, ("initHardware: bInit=%s\n", bInit ? "TRUE" : "FALSE"));

  pRegs = pDevCtx->hardware.bridge.pPci9656;
  hBridge = pDevCtx->hardware.hBridge;

  if (bInit) {
    pDevCtx->info.model = CoreModelAdmxrc2;

    /* Enable Direct Master access */
    dfPciMemWrite32(hBridge, &pRegs->DMRR, 0xFF000000U);
    dfPciMemWrite32(hBridge, &pRegs->DMLBAM, 0xFF000000U);
    dfPciMemWrite32(hBridge, &pRegs->DMLBAI, 0x0U);
    dfPciMemWrite32(hBridge, &pRegs->DMPBAM, 0x1U);

    /* Ensure use of PCI Read Multiple commands when a bus master, for better performance */
    val32 = dfPciMemRead32(hBridge, &pRegs->CNTRL);
    val32 &= ~0x0f0fU;
    val32 |= 0x0c0cU;
    dfPciMemWrite32(hBridge, &pRegs->CNTRL, val32);

    if (pDevCtx->info.bootstrap.flashWidth != 0) {
      if (flashCfiIdentify(pDevCtx, 0, 0)) {
        pDevCtx->info.flash.bPresent = TRUE;
        pDevCtx->info.flash.targetArea.start = 0;
        pDevCtx->info.flash.targetArea.length = pDevCtx->info.flash.totalSize;
      } else {
        dfDebugPrint(0, ("*** initHardware: failed to identify Flash\n"));
      }
    }

    status = readWriteVpd(pDevCtx, FALSE, VPD_OFFSET, sizeof(vpdRaw), &vpdRaw);
    if (CoreVpdSuccess == status) {
      pDevCtx->info.vpd.bValid = TRUE;
      pVpd = &pDevCtx->info.vpd.pData->admxrc2.rev0;
      pVpd->model = dfLe16ToCpu(vpdRaw.model);
      pVpd->modifications = dfLe16ToCpu(vpdRaw.modifications);
      pVpd->refClock = dfLe16ToCpu(vpdRaw.refClock);
      pVpd->flashType = dfLe16ToCpu(vpdRaw.flashType);
      pVpd->zbt01 = dfLe8ToCpu(vpdRaw.zbt01);
      pVpd->zbt23 = dfLe8ToCpu(vpdRaw.zbt23);
      pVpd->zbt45 = dfLe8ToCpu(vpdRaw.zbt45);
      pVpd->cardId = dfLe16ToCpu(vpdRaw.cardId);
      pVpd->fpgaType = dfLe16ToCpu(vpdRaw.fpgaType);
      pVpd->serialNumber = dfLe32ToCpu(vpdRaw.serialNumber);
      pVpd->cpldRev = dfLe32ToCpu(vpdRaw.cpldRev);
      pVpd->pcbRev = dfLe32ToCpu(vpdRaw.pcbRev);
    } else {
      dfDebugPrint(0, ("*** initHardware: failed to read VPD from EEPROM, status=0x%lx\n", (unsigned long)status));
    }

    /* Find base addresses and sizes of local bus windows */
    pDevCtx->info.window[0].busBase = pDevCtx->rawResources.variant.pci.memBar[1].base;
    pDevCtx->info.window[0].busSize = pDevCtx->rawResources.variant.pci.memBar[1].size;
    ls0b = dfPciMemRead32(hBridge, &pRegs->LAS0BA);
    ls0r = dfPciMemRead32(hBridge, &pRegs->LAS0RR);
    ls0r &= 0xfffffff0U;
    ls0b &= ls0r;
    pDevCtx->info.window[0].localBase = ls0b;
    pDevCtx->info.window[0].localSize = ~ls0r + 1;
    pDevCtx->info.window[0].translatedBase = pDevCtx->translatedResources.variant.pci.memBar[1].base;
    pDevCtx->info.window[0].translatedSize = pDevCtx->translatedResources.variant.pci.memBar[1].size;
    pDevCtx->info.window[0].pKernelBase = pDevCtx->hardware.pBridged;
    pDevCtx->info.window[0].kernelSize = (size_t)pDevCtx->translatedResources.variant.pci.memBar[1].size;
    pDevCtx->info.window[1].busBase = pDevCtx->rawResources.variant.pci.memBar[2].base;
    pDevCtx->info.window[1].busSize = pDevCtx->rawResources.variant.pci.memBar[2].size;
    ls1b = dfPciMemRead32(hBridge, &pRegs->LAS1BA);
    ls1r = dfPciMemRead32(hBridge, &pRegs->LAS1RR);
    ls1r &= 0xfffffff0U;
    ls1b &= ls1r;
    pDevCtx->info.window[1].localBase = ls1b;
    pDevCtx->info.window[1].localSize = ~ls1r + 1;
    pDevCtx->info.window[1].translatedBase = pDevCtx->translatedResources.variant.pci.memBar[2].base;
    pDevCtx->info.window[1].translatedSize = pDevCtx->translatedResources.variant.pci.memBar[2].size;
    pDevCtx->info.window[1].pKernelBase = pDevCtx->hardware.model.pAdmxrc2;
    pDevCtx->info.window[1].kernelSize = (size_t)pDevCtx->translatedResources.variant.pci.memBar[2].size;
    pDevCtx->info.window[2].busBase = pDevCtx->rawResources.variant.pci.memBar[0].base;
    pDevCtx->info.window[2].busSize = pDevCtx->rawResources.variant.pci.memBar[0].size;
    pDevCtx->info.window[2].localBase = 0;
    pDevCtx->info.window[2].localSize = 0;
    pDevCtx->info.window[2].translatedBase = pDevCtx->translatedResources.variant.pci.memBar[0].base;
    pDevCtx->info.window[2].translatedSize = pDevCtx->translatedResources.variant.pci.memBar[0].size;
    pDevCtx->info.window[2].pKernelBase = pDevCtx->hardware.bridge.pGeneric;
    pDevCtx->info.window[2].kernelSize = (size_t)pDevCtx->translatedResources.variant.pci.memBar[0].size;
  } else {
    /* Nothing to do */
  }

  dfDebugPrint(1, ("initHardware: done\n"));
  return TRUE;
}

static boolean_t
isr(
  DfInterruptObject* pInterruptObj,
  void* pInterruptContext)
{
  Adb3CoreDeviceContext* pDevCtx = (Adb3CoreDeviceContext*)pInterruptContext;
  Pci9656Registers* pRegs;
  DfMemoryHandle hBridge;
  ModelRegsAdmxrc2* pModelRegs;
  DfMemoryHandle hModel;
  uint32_t intcsr, clearMask;
  uint8_t istat;

  dfDebugPrint(8, ("isr: entered\n"));

  pRegs = pDevCtx->hardware.bridge.pPci9656;
  hBridge = pDevCtx->hardware.hBridge;
  pModelRegs = pDevCtx->hardware.model.pAdmxrc2;
  hModel = pDevCtx->hardware.hModel;

  /* Sample the currently active interrupts, and mask the value with the interrupts we want */
  intcsr = dfPciMemRead32(hBridge, &pRegs->INTCSR) & pDevCtx->hardware.shadow.bridge.pci9656.intcsrMask;

  /* Clear the interrupts we saw */
  dfPciMemWrite32(hBridge, &pRegs->INTCSR, intcsr | pDevCtx->hardware.shadow.bridge.pci9656.intcsr);

  if (intcsr & PLX_INTCSR_LOC_INT) {
    dfDebugPrint(9, ("isr: MODEL interrupt\n"));
    istat = dfPciMemRead8(hModel, &pModelRegs->icon) & pDevCtx->hardware.shadow.model.admxrc2.ienab;
    /* Clear the model-specific interrupts that we saw */
    dfPciMemWrite8(hModel, &pModelRegs->icon, istat);
    if (istat & ADMXRC2_ICON_FPGA) {
      dfDebugPrint(9, ("isr: target FPGA interrupt\n"));
      dfDpcSchedule(&pDevCtx->interrupt.fpga[0].dpc, (void*)(uintptr_t)0);
    }
  }

  if (intcsr & (PLX_INTCSR_DMA0_INT | PLX_INTCSR_DMA1_INT)) {
    clearMask = 0x0U;
    if (intcsr & PLX_INTCSR_DMA0_INT) {
      dfDebugPrint(9, ("isr: DMA0 interrupt\n"));
      clearMask |= 0x0008U;
    }
    if (intcsr & PLX_INTCSR_DMA1_INT) {
      dfDebugPrint(9, ("isr: DMA1 interrupt\n"));
      clearMask |= 0x0800U;
    }
    dfPciMemWrite16(hBridge, &pRegs->DMACSR, clearMask | pDevCtx->hardware.shadow.bridge.pci9656.dmacsr);
    if (intcsr & PLX_INTCSR_DMA0_INT) {
      dfDpcSchedule(&pDevCtx->dma.channel[0].completionDpc, (void*)(uintptr_t)0);
    }
    if (intcsr & PLX_INTCSR_DMA1_INT) {
      dfDpcSchedule(&pDevCtx->dma.channel[1].completionDpc, (void*)(uintptr_t)1);
    }
  }

  return intcsr ? TRUE : FALSE;
}

static boolean_t
mapBusResources(
  Adb3CoreDeviceContext* pDevCtx,
  DfBusResources* pRaw,
  DfBusResources* pTranslated,
  boolean_t bUnmap)
{
  dfDebugPrint(2, ("mapBusResources: bUnmap=%s\n", bUnmap ? "TRUE" : "FALSE"));
  
  if (bUnmap) {
    if (NULL != pDevCtx->hardware.bridge.pGeneric) {
      if (!dfIoSpaceUnmap(pDevCtx->pDevObj, pDevCtx->hardware.bridge.pGeneric)) {
        dfDebugPrint(0, ("*** mapBusResources: failed to unmap PCI9656\n"));
      }
    }
    if (NULL != pDevCtx->hardware.model.pGeneric) {
      if (!dfIoSpaceUnmap(pDevCtx->pDevObj, pDevCtx->hardware.model.pGeneric)) {
        dfDebugPrint(0, ("*** mapBusResources: failed to unmap CPLD\n"));
      }
    }
    if (NULL != pDevCtx->hardware.pBridged) {
      if (!dfIoSpaceUnmap(pDevCtx->pDevObj, pDevCtx->hardware.pBridged)) {
        dfDebugPrint(0, ("*** mapBusResources: failed to unmap FPGA space\n"));
      }
    }
    dfDebugPrint(2, ("mapBusResources: unmapped\n"));
  } else {
    pDevCtx->hardware.bridge.pGeneric = dfIoSpaceMap(pDevCtx->pDevObj, pTranslated, 0, 0, 0x200U, &pDevCtx->hardware.hBridge);
    if (NULL == pDevCtx->hardware.bridge.pGeneric) {
      dfDebugPrint(0, ("*** mapBusResources: failed to map PCI9656\n"));
      return FALSE;
    } else {
      dfDebugPrint(2, ("mapBusResources: mapped PCI9656 @ 0x%08lx_%08lx to %p\n",
        dfSplitUint64(pTranslated->variant.pci.memBar[0].base), pDevCtx->hardware.bridge.pGeneric));
    }
    pDevCtx->hardware.model.pGeneric = dfIoSpaceMap(pDevCtx->pDevObj, pTranslated, 2, 0, 0x400000U, &pDevCtx->hardware.hModel);
    if (NULL == pDevCtx->hardware.model.pGeneric) {
      dfDebugPrint(0, ("*** mapBusResources: failed to map CPLD\n"));
      return FALSE;
    } else {
      dfDebugPrint(2, ("mapBusResources: mapped CPLD @ 0x%08lx_%08lx to %p\n",
      dfSplitUint64(pTranslated->variant.pci.memBar[2].base), pDevCtx->hardware.model.pGeneric));
    }
    pDevCtx->hardware.pBridged = dfIoSpaceMap(pDevCtx->pDevObj, pTranslated, 1, 0, pTranslated->variant.pci.memBar[1].size, &pDevCtx->hardware.hBridged);
    if (NULL == pDevCtx->hardware.pBridged) {
      dfDebugPrint(0, ("*** mapBusResources: failed to map FPGA space\n"));
      return FALSE;
    } else {
      dfDebugPrint(2, ("mapBusResources: mapped FPGA space @ 0x%08lx_%08lx to %p\n",
        dfSplitUint64(pTranslated->variant.pci.memBar[1].base), pDevCtx->hardware.pBridged));
    }
    dfDebugPrint(2, ("mapBusResources: mapped OK\n"));
  }
  
  return TRUE;
}

static CoreVpdStatus
readWriteVpd(
  Adb3CoreDeviceContext* pDevCtx,
  boolean_t bWrite,
  size_t address,
  size_t length,
  void* pData)
{
  if (CHECK_BOUNDS_OFFSET(address, length, pDevCtx->info.bootstrap.vpdOffset, pDevCtx->info.bootstrap.vpdLength)) {
    return CoreVpdInvalidRegion;
  }
  if (bWrite) {
    return write93x6Buffer(pDevCtx, TRUE, addressLength9356, instructionLength9356, (uint32_t)address, (uint32_t)length, pData);
  } else {
    return read93x6Buffer(pDevCtx, TRUE, addressLength9356, instructionLength9356, (uint32_t)address, (uint32_t)length, pData);
  }
}

static boolean_t
validateBusResources(
  Adb3CoreDeviceContext* pDevCtx,
  DfBusResources* pRaw,
  DfBusResources* pTranslated)
{
  if (pTranslated->variant.pci.numMemBar < 3) {
    dfDebugPrint(0, ("*** validateBusResources: less than 3 memory BARs\n"));
    return FALSE;
  }
  if (pTranslated->variant.pci.memBar[0].size < 0x200U) {
    dfDebugPrint(0, ("*** validateBusResources: BAR0 smaller than 0x200 bytes\n"));
    return FALSE;
  }
  if (pTranslated->variant.pci.memBar[1].size < 0x400000U) {
    dfDebugPrint(0, ("*** validateBusResources: BAR1 smaller than 0x400000 bytes\n"));
    return FALSE;
  }
  if (pTranslated->variant.pci.memBar[2].size < 0x400000U) {
    dfDebugPrint(0, ("*** validateBusResources: BAR2 smaller than 0x400000 bytes\n"));
    return FALSE;
  }
  if (pTranslated->variant.pci.numIrq != 1) {
    dfDebugPrint(0, ("*** validateBusResources: not exactly 1 interrupt\n"));
    return FALSE;
  }
  
  dfDebugPrint(2, ("validateBusResources: OK\n"));
  
  return TRUE;
}

void
adb3CoreBootstrapAdmxrc2(
  Adb3CoreDeviceContext* pDevCtx)
{
  dfDebugPrint(1, ("adb3CoreBootstrapAdmxrc2: entered\n"));

  pDevCtx->methods.pClockProgram = clockProgram;
  pDevCtx->methods.pClockWord = clockWord;
  pDevCtx->methods.pDmaList = dmaList;
  pDevCtx->methods.pDmaTransfer = dmaTransfer;
  pDevCtx->methods.pFlash = flash;
  pDevCtx->methods.pEnableInterrupts = enableInterrupts;
  pDevCtx->methods.pFpgaControl = fpgaControl;
  pDevCtx->methods.pFpgaSelectMap = fpgaSelectMap;
  pDevCtx->methods.pInitHardware = initHardware;
  pDevCtx->methods.pIsr = isr;
  pDevCtx->methods.pMapBusResources = mapBusResources;
  pDevCtx->methods.pReadWriteVpd = readWriteVpd;
  pDevCtx->methods.pValidateBusResources = validateBusResources;

  pDevCtx->info.bootstrap.numClockGen = 2;
  pDevCtx->info.bootstrap.numDmaChannel = 2;
  pDevCtx->info.bootstrap.numIoModule = 1;
  pDevCtx->info.bootstrap.numSensor = 0;
  pDevCtx->info.bootstrap.numTargetFpga = 1;
  pDevCtx->info.bootstrap.numWindow = 3;
  pDevCtx->info.bootstrap.bVpdInFlash = FALSE;
  pDevCtx->info.bootstrap.vpdLength = eepromSize9356;
  pDevCtx->info.bootstrap.vpdOffset = 0;
  pDevCtx->info.bootstrap.vpdBufferLength = sizeof(VpdAdmxrc2Rev0);
  pDevCtx->info.bootstrap.flashDeviceWidth = 2;
  pDevCtx->info.bootstrap.flashWidth = 1;
  pDevCtx->info.bootstrap.b64BitDmaAddress = FALSE;
  pDevCtx->info.bootstrap.dmaDescriptorSize = 16;
  pDevCtx->info.bootstrap.dmaTransferMaxSize = 0x800000U;
}
